@@ -34,7 +34,6 @@
 #include <telepathy-glib/account-manager.h>
 #include <telepathy-glib/util.h>
 
-#include <libempathy/empathy-call-factory.h>
 #include <libempathy/empathy-tp-contact-factory.h>
 #include <libempathy/empathy-contact-list.h>
 #include <libempathy/empathy-contact-groups.h>
@@ -387,8 +386,8 @@ contact_list_view_contact_drag_received (GtkWidget         *view,
 {
 	EmpathyContactListViewPriv *priv;
 	TpAccountManager           *account_manager;
-        TpConnection               *connection;
-	TpAccount                  *account;
+	TpConnection               *connection = NULL;
+	TpAccount                  *account = NULL;
 	DndGetContactData          *data;
 	GtkTreePath                *source_path;
 	const gchar   *sel_data;
@@ -804,6 +803,17 @@ typedef struct {
 	guint32                 time;
 } MenuPopupData;
 
+static void
+menu_deactivate_cb (GtkMenuShell *menushell,
+		    gpointer user_data)
+{
+	/* FIXME: we shouldn't have to disconnec the signal (bgo #641327) */
+	g_signal_handlers_disconnect_by_func (menushell,
+		menu_deactivate_cb, user_data);
+
+	gtk_menu_detach (GTK_MENU (menushell));
+}
+
 static gboolean
 contact_list_view_popup_menu_idle_cb (gpointer user_data)
 {
@@ -822,8 +832,15 @@ contact_list_view_popup_menu_idle_cb (gpointer user_data)
 		gtk_menu_popup (GTK_MENU (menu),
 				NULL, NULL, NULL, NULL,
 				data->button, data->time);
-		g_object_ref_sink (menu);
-		g_object_unref (menu);
+
+		/* menu is initially unowned but gtk_menu_attach_to_widget () taked its
+		 * floating ref. We can either wait that the treeview releases its ref
+		 * when it will be destroyed (when leaving Empathy) or explicitely
+		 * detach the menu when it's not displayed any more.
+		 * We go for the latter as we don't want to keep useless menus in memory
+		 * during the whole lifetime of Empathy. */
+		g_signal_connect (menu, "deactivate", G_CALLBACK (menu_deactivate_cb),
+			NULL);
 	}
 
 	g_slice_free (MenuPopupData, data);
@@ -921,7 +938,7 @@ contact_list_view_call_activated_cb (
 
 	event = (GdkEventButton *) gtk_get_current_event ();
 
-	menu = gtk_menu_new ();
+	menu = empathy_context_menu_new (GTK_WIDGET (view));
 	shell = GTK_MENU_SHELL (menu);
 
 	/* audio */
@@ -934,13 +951,9 @@ contact_list_view_call_activated_cb (
 	gtk_menu_shell_append (shell, item);
 	gtk_widget_show (item);
 
-	gtk_menu_attach_to_widget (GTK_MENU (menu),
-				   GTK_WIDGET (view), NULL);
 	gtk_widget_show (menu);
 	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
 			event->button, event->time);
-	g_object_ref_sink (menu);
-	g_object_unref (menu);
 
 	g_object_unref (contact);
 }
@@ -951,29 +964,28 @@ contact_list_view_cell_set_background (EmpathyContactListView *view,
 				       gboolean                is_group,
 				       gboolean                is_active)
 {
-	GdkColor  color;
-	GtkStyle *style;
-
-	style = gtk_widget_get_style (GTK_WIDGET (view));
-
 	if (!is_group && is_active) {
-		color = style->bg[GTK_STATE_SELECTED];
+		GdkRGBA color;
+		GtkStyleContext *style;
+
+		style = gtk_widget_get_style_context (GTK_WIDGET (view));
+
+		gtk_style_context_get_background_color (style, GTK_STATE_FLAG_SELECTED,
+        &color);
 
 		/* Here we take the current theme colour and add it to
 		 * the colour for white and average the two. This
 		 * gives a colour which is inline with the theme but
 		 * slightly whiter.
 		 */
-		color.red = (color.red + (style->white).red) / 2;
-		color.green = (color.green + (style->white).green) / 2;
-		color.blue = (color.blue + (style->white).blue) / 2;
+		empathy_make_color_whiter (&color);
 
 		g_object_set (cell,
-			      "cell-background-gdk", &color,
+			      "cell-background-rgba", &color,
 			      NULL);
 	} else {
 		g_object_set (cell,
-			      "cell-background-gdk", NULL,
+			      "cell-background-rgba", NULL,
 			      NULL);
 	}
 }
