@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2004-2007 Imendio AB
- * Copyright (C) 2007-2009 Collabora Ltd.
+ * Copyright (C) 2007-2010 Collabora Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -32,6 +32,7 @@
 
 #include <telepathy-glib/account-manager.h>
 #include <telepathy-glib/interfaces.h>
+#include <telepathy-glib/simple-observer.h>
 #include <telepathy-glib/util.h>
 
 #include "empathy-tp-chat.h"
@@ -47,6 +48,15 @@
 
 static EmpathyChatroomManager *chatroom_manager_singleton = NULL;
 
+static void observe_channels_cb (TpSimpleObserver *observer,
+    TpAccount *account,
+    TpConnection *connection,
+    GList *channels,
+    TpChannelDispatchOperation *dispatch_operation,
+    GList *requests,
+    TpObserveChannelsContext *context,
+    gpointer user_data);
+
 #define GET_PRIV(obj) EMPATHY_GET_PRIV (obj, EmpathyChatroomManager)
 typedef struct
 {
@@ -57,6 +67,10 @@ typedef struct
   /* source id of the autosave timer */
   gint save_timer_id;
   gboolean ready;
+  GFileMonitor *monitor;
+  gboolean writing;
+
+  TpBaseClient *observer;
 } EmpathyChatroomManagerPriv;
 
 enum {
@@ -84,56 +98,59 @@ G_DEFINE_TYPE (EmpathyChatroomManager, empathy_chatroom_manager, G_TYPE_OBJECT);
 static gboolean
 chatroom_manager_file_save (EmpathyChatroomManager *manager)
 {
-	EmpathyChatroomManagerPriv *priv;
-	xmlDocPtr                  doc;
-	xmlNodePtr                 root;
-	GList                     *l;
+  EmpathyChatroomManagerPriv *priv;
+  xmlDocPtr doc;
+  xmlNodePtr root;
+  GList *l;
 
-	priv = GET_PRIV (manager);
+  priv = GET_PRIV (manager);
 
-	doc = xmlNewDoc ((const xmlChar *) "1.0");
-	root = xmlNewNode (NULL, (const xmlChar *) "chatrooms");
-	xmlDocSetRootElement (doc, root);
+  priv->writing = TRUE;
 
-	for (l = priv->chatrooms; l; l = l->next) {
-		EmpathyChatroom *chatroom;
-		xmlNodePtr       node;
-		const gchar     *account_id;
+  doc = xmlNewDoc ((const xmlChar *) "1.0");
+  root = xmlNewNode (NULL, (const xmlChar *) "chatrooms");
+  xmlDocSetRootElement (doc, root);
 
-		chatroom = l->data;
+  for (l = priv->chatrooms; l; l = l->next)
+    {
+      EmpathyChatroom *chatroom;
+      xmlNodePtr       node;
+      const gchar     *account_id;
 
-		if (!empathy_chatroom_is_favorite (chatroom)) {
-			continue;
-		}
+      chatroom = l->data;
 
-                account_id = tp_proxy_get_object_path (
-		  empathy_chatroom_get_account (chatroom));
+      if (!empathy_chatroom_is_favorite (chatroom))
+        continue;
 
-		node = xmlNewChild (root, NULL, (const xmlChar *) "chatroom", NULL);
-		xmlNewTextChild (node, NULL, (const xmlChar *) "name",
-			(const xmlChar *) empathy_chatroom_get_name (chatroom));
-		xmlNewTextChild (node, NULL, (const xmlChar *) "room",
-			(const xmlChar *) empathy_chatroom_get_room (chatroom));
-		xmlNewTextChild (node, NULL, (const xmlChar *) "account",
-			(const xmlChar *) account_id);
-		xmlNewTextChild (node, NULL, (const xmlChar *) "auto_connect",
-			empathy_chatroom_get_auto_connect (chatroom) ?
-			(const xmlChar *) "yes" : (const xmlChar *) "no");
-		xmlNewTextChild (node, NULL, (const xmlChar *) "always_urgent",
-			empathy_chatroom_is_always_urgent (chatroom) ?
-			(const xmlChar *) "yes" : (const xmlChar *) "no");
-	}
+      account_id = tp_proxy_get_object_path (empathy_chatroom_get_account (
+            chatroom));
 
-	/* Make sure the XML is indented properly */
-	xmlIndentTreeOutput = 1;
+      node = xmlNewChild (root, NULL, (const xmlChar *) "chatroom", NULL);
+      xmlNewTextChild (node, NULL, (const xmlChar *) "name",
+        (const xmlChar *) empathy_chatroom_get_name (chatroom));
+      xmlNewTextChild (node, NULL, (const xmlChar *) "room",
+        (const xmlChar *) empathy_chatroom_get_room (chatroom));
+      xmlNewTextChild (node, NULL, (const xmlChar *) "account",
+        (const xmlChar *) account_id);
+      xmlNewTextChild (node, NULL, (const xmlChar *) "auto_connect",
+        empathy_chatroom_get_auto_connect (chatroom) ?
+        (const xmlChar *) "yes" : (const xmlChar *) "no");
+      xmlNewTextChild (node, NULL, (const xmlChar *) "always_urgent",
+        empathy_chatroom_is_always_urgent (chatroom) ?
+        (const xmlChar *) "yes" : (const xmlChar *) "no");
+    }
 
-	DEBUG ("Saving file:'%s'", priv->file);
-	xmlSaveFormatFileEnc (priv->file, doc, "utf-8", 1);
-	xmlFreeDoc (doc);
+  /* Make sure the XML is indented properly */
+  xmlIndentTreeOutput = 1;
 
-	xmlMemoryDump ();
+  DEBUG ("Saving file:'%s'", priv->file);
+  xmlSaveFormatFileEnc (priv->file, doc, "utf-8", 1);
+  xmlFreeDoc (doc);
 
-	return TRUE;
+  xmlMemoryDump ();
+
+  priv->writing = FALSE;
+  return TRUE;
 }
 
 static gboolean
@@ -153,9 +170,7 @@ reset_save_timeout (EmpathyChatroomManager *self)
   EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
 
   if (priv->save_timer_id > 0)
-    {
-      g_source_remove (priv->save_timer_id);
-    }
+    g_source_remove (priv->save_timer_id);
 
   priv->save_timer_id = g_timeout_add_seconds (SAVE_TIMER,
       (GSourceFunc) save_timeout, self);
@@ -163,15 +178,15 @@ reset_save_timeout (EmpathyChatroomManager *self)
 
 static void
 chatroom_changed_cb (EmpathyChatroom *chatroom,
-                     GParamSpec *spec,
-                     EmpathyChatroomManager *self)
+    GParamSpec *spec,
+    EmpathyChatroomManager *self)
 {
   reset_save_timeout (self);
 }
 
 static void
 add_chatroom (EmpathyChatroomManager *self,
-              EmpathyChatroom *chatroom)
+    EmpathyChatroom *chatroom)
 {
   EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
 
@@ -183,157 +198,166 @@ add_chatroom (EmpathyChatroomManager *self,
 
 static void
 chatroom_manager_parse_chatroom (EmpathyChatroomManager *manager,
-				 xmlNodePtr             node)
+    xmlNodePtr node)
 {
-	EmpathyChatroomManagerPriv *priv;
-	EmpathyChatroom            *chatroom;
-	TpAccount                  *account;
-	xmlNodePtr                 child;
-	gchar                     *str;
-	gchar                     *name;
-	gchar                     *room;
-	gchar                     *account_id;
-	gboolean                   auto_connect;
-	gboolean                   always_urgent;
+  EmpathyChatroomManagerPriv *priv;
+  EmpathyChatroom *chatroom;
+  TpAccount *account;
+  xmlNodePtr child;
+  gchar *str;
+  gchar *name;
+  gchar *room;
+  gchar *account_id;
+  gboolean auto_connect;
+  gboolean always_urgent;
 
-	priv = GET_PRIV (manager);
+  priv = GET_PRIV (manager);
 
-	/* default values. */
-	name = NULL;
-	room = NULL;
-	auto_connect = TRUE;
-	always_urgent = FALSE;
-	account_id = NULL;
+  /* default values. */
+  name = NULL;
+  room = NULL;
+  auto_connect = TRUE;
+  always_urgent = FALSE;
+  account_id = NULL;
 
-	for (child = node->children; child; child = child->next) {
-		gchar *tag;
+  for (child = node->children; child; child = child->next)
+    {
+      gchar *tag;
 
-		if (xmlNodeIsText (child)) {
-			continue;
-		}
+      if (xmlNodeIsText (child))
+        continue;
 
-		tag = (gchar *) child->name;
-		str = (gchar *) xmlNodeGetContent (child);
+      tag = (gchar *) child->name;
+      str = (gchar *) xmlNodeGetContent (child);
 
-		if (strcmp (tag, "name") == 0) {
-			name = g_strdup (str);
-		}
-		else if (strcmp (tag, "room") == 0) {
-			room = g_strdup (str);
-		}
-		else if (strcmp (tag, "auto_connect") == 0) {
-			if (strcmp (str, "yes") == 0) {
-				auto_connect = TRUE;
-			} else {
-				auto_connect = FALSE;
-			}
-		}
-		else if (!tp_strdiff (tag, "always_urgent")) {
-			if (strcmp (str, "yes") == 0) {
-				always_urgent = TRUE;
-			} else {
-				always_urgent = FALSE;
-			}
-		}
-		else if (strcmp (tag, "account") == 0) {
-			account_id = g_strdup (str);
-		}
+      if (strcmp (tag, "name") == 0)
+        {
+          name = g_strdup (str);
+        }
+      else if (strcmp (tag, "room") == 0)
+        {
+          room = g_strdup (str);
+        }
+      else if (strcmp (tag, "auto_connect") == 0)
+        {
+          if (strcmp (str, "yes") == 0)
+            auto_connect = TRUE;
+          else
+            auto_connect = FALSE;
+        }
+      else if (!tp_strdiff (tag, "always_urgent"))
+        {
+          if (strcmp (str, "yes") == 0)
+            always_urgent = TRUE;
+          else
+            always_urgent = FALSE;
+        }
+      else if (strcmp (tag, "account") == 0)
+        {
+          account_id = g_strdup (str);
+        }
 
-		xmlFree (str);
-	}
+      xmlFree (str);
+    }
 
-	account = tp_account_manager_ensure_account (priv->account_manager,
-		account_id);
-	if (!account) {
-		g_free (name);
-		g_free (room);
-		g_free (account_id);
-		return;
-	}
+  account = tp_account_manager_ensure_account (priv->account_manager,
+    account_id);
+  if (account == NULL)
+    {
+      g_free (name);
+      g_free (room);
+      g_free (account_id);
+      return;
+    }
 
-	chatroom = empathy_chatroom_new_full (account, room, name, auto_connect);
-	empathy_chatroom_set_favorite (chatroom, TRUE);
-	empathy_chatroom_set_always_urgent (chatroom, always_urgent);
-	add_chatroom (manager, chatroom);
-	g_signal_emit (manager, signals[CHATROOM_ADDED], 0, chatroom);
+  chatroom = empathy_chatroom_new_full (account, room, name, auto_connect);
+  empathy_chatroom_set_favorite (chatroom, TRUE);
+  empathy_chatroom_set_always_urgent (chatroom, always_urgent);
+  add_chatroom (manager, chatroom);
+  g_signal_emit (manager, signals[CHATROOM_ADDED], 0, chatroom);
 
-	g_free (name);
-	g_free (room);
-	g_free (account_id);
+  g_free (name);
+  g_free (room);
+  g_free (account_id);
+  g_object_unref (chatroom);
 }
 
 static gboolean
 chatroom_manager_file_parse (EmpathyChatroomManager *manager,
-			     const gchar           *filename)
+    const gchar *filename)
 {
-	EmpathyChatroomManagerPriv *priv;
-	xmlParserCtxtPtr           ctxt;
-	xmlDocPtr                  doc;
-	xmlNodePtr                 chatrooms;
-	xmlNodePtr                 node;
+  EmpathyChatroomManagerPriv *priv;
+  xmlParserCtxtPtr ctxt;
+  xmlDocPtr doc;
+  xmlNodePtr chatrooms;
+  xmlNodePtr node;
 
-	priv = GET_PRIV (manager);
+  priv = GET_PRIV (manager);
 
-	DEBUG ("Attempting to parse file:'%s'...", filename);
+  DEBUG ("Attempting to parse file:'%s'...", filename);
 
-	ctxt = xmlNewParserCtxt ();
+  ctxt = xmlNewParserCtxt ();
 
-	/* Parse and validate the file. */
-	doc = xmlCtxtReadFile (ctxt, filename, NULL, 0);
-	if (!doc) {
-		g_warning ("Failed to parse file:'%s'", filename);
-		xmlFreeParserCtxt (ctxt);
-		return FALSE;
-	}
+  /* Parse and validate the file. */
+  doc = xmlCtxtReadFile (ctxt, filename, NULL, 0);
+  if (doc == NULL)
+    {
+      g_warning ("Failed to parse file:'%s'", filename);
+      xmlFreeParserCtxt (ctxt);
+      return FALSE;
+    }
 
-	if (!empathy_xml_validate (doc, CHATROOMS_DTD_FILENAME)) {
-		g_warning ("Failed to validate file:'%s'", filename);
-		xmlFreeDoc (doc);
-		xmlFreeParserCtxt (ctxt);
-		return FALSE;
-	}
+  if (!empathy_xml_validate (doc, CHATROOMS_DTD_FILENAME))
+    {
+      g_warning ("Failed to validate file:'%s'", filename);
+      xmlFreeDoc (doc);
+      xmlFreeParserCtxt (ctxt);
+      return FALSE;
+    }
 
-	/* The root node, chatrooms. */
-	chatrooms = xmlDocGetRootElement (doc);
+  /* The root node, chatrooms. */
+  chatrooms = xmlDocGetRootElement (doc);
 
-	for (node = chatrooms->children; node; node = node->next) {
-		if (strcmp ((gchar *) node->name, "chatroom") == 0) {
-			chatroom_manager_parse_chatroom (manager, node);
-		}
-	}
+  for (node = chatrooms->children; node; node = node->next)
+    {
+      if (strcmp ((gchar *) node->name, "chatroom") == 0)
+        chatroom_manager_parse_chatroom (manager, node);
+    }
 
-	DEBUG ("Parsed %d chatrooms", g_list_length (priv->chatrooms));
+  DEBUG ("Parsed %d chatrooms", g_list_length (priv->chatrooms));
 
-	xmlFreeDoc (doc);
-	xmlFreeParserCtxt (ctxt);
+  xmlFreeDoc (doc);
+  xmlFreeParserCtxt (ctxt);
 
-	return TRUE;
+  return TRUE;
 }
 
 static gboolean
 chatroom_manager_get_all (EmpathyChatroomManager *manager)
 {
-	EmpathyChatroomManagerPriv *priv;
+  EmpathyChatroomManagerPriv *priv;
 
-	priv = GET_PRIV (manager);
+  priv = GET_PRIV (manager);
 
-	/* read file in */
-	if (g_file_test (priv->file, G_FILE_TEST_EXISTS) &&
-	    !chatroom_manager_file_parse (manager, priv->file)) {
-		return FALSE;
-	}
+  /* read file in */
+  if (g_file_test (priv->file, G_FILE_TEST_EXISTS) &&
+      !chatroom_manager_file_parse (manager, priv->file))
+    return FALSE;
 
-        priv->ready = TRUE;
-        g_object_notify (G_OBJECT (manager), "ready");
+  if (!priv->ready)
+    {
+      priv->ready = TRUE;
+      g_object_notify (G_OBJECT (manager), "ready");
+    }
 
-	return TRUE;
+  return TRUE;
 }
 
 static void
 empathy_chatroom_manager_get_property (GObject *object,
-                                       guint property_id,
-                                       GValue *value,
-                                       GParamSpec *pspec)
+    guint property_id,
+    GValue *value,
+    GParamSpec *pspec)
 {
   EmpathyChatroomManager *self = EMPATHY_CHATROOM_MANAGER (object);
   EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
@@ -354,9 +378,9 @@ empathy_chatroom_manager_get_property (GObject *object,
 
 static void
 empathy_chatroom_manager_set_property (GObject *object,
-                                       guint property_id,
-                                       const GValue *value,
-                                       GParamSpec *pspec)
+    guint property_id,
+    const GValue *value,
+    GParamSpec *pspec)
 {
   EmpathyChatroomManager *self = EMPATHY_CHATROOM_MANAGER (object);
   EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
@@ -373,12 +397,45 @@ empathy_chatroom_manager_set_property (GObject *object,
     }
 }
 
+static void
+chatroom_manager_dispose (GObject *object)
+{
+  EmpathyChatroomManagerPriv *priv;
+
+  priv = GET_PRIV (object);
+
+  tp_clear_object (&priv->observer);
+  tp_clear_object (&priv->monitor);
+
+  (G_OBJECT_CLASS (empathy_chatroom_manager_parent_class)->dispose) (object);
+}
+
+static void
+clear_chatrooms (EmpathyChatroomManager *self)
+{
+  EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
+  GList *l;
+
+  for (l = priv->chatrooms; l != NULL; l = g_list_next (l))
+    {
+      EmpathyChatroom *chatroom = l->data;
+
+      g_signal_handlers_disconnect_by_func (chatroom, chatroom_changed_cb,
+          self);
+      g_signal_emit (self, signals[CHATROOM_REMOVED], 0, chatroom);
+
+      g_object_unref (chatroom);
+    }
+
+  g_list_free (priv->chatrooms);
+  priv->chatrooms = NULL;
+}
+
 static void
 chatroom_manager_finalize (GObject *object)
 {
   EmpathyChatroomManager *self = EMPATHY_CHATROOM_MANAGER (object);
   EmpathyChatroomManagerPriv *priv;
-  GList *l;
 
   priv = GET_PRIV (object);
 
@@ -392,30 +449,45 @@ chatroom_manager_finalize (GObject *object)
       chatroom_manager_file_save (self);
     }
 
-  for (l = priv->chatrooms; l != NULL; l = g_list_next (l))
-    {
-      EmpathyChatroom *chatroom = l->data;
+  clear_chatrooms (self);
 
-      g_signal_handlers_disconnect_by_func (chatroom, chatroom_changed_cb,
-          self);
-
-      g_object_unref (chatroom);
-    }
-
-  g_list_free (priv->chatrooms);
   g_free (priv->file);
 
   (G_OBJECT_CLASS (empathy_chatroom_manager_parent_class)->finalize) (object);
 }
 
+static void
+file_changed_cb (GFileMonitor *monitor,
+    GFile *file,
+    GFile *other_file,
+    GFileMonitorEvent event_type,
+    gpointer user_data)
+{
+  EmpathyChatroomManager *self = user_data;
+  EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
+
+  if (event_type != G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT)
+    return;
+
+  if (priv->writing)
+    return;
+
+  DEBUG ("chatrooms file changed; reloading list");
+
+  clear_chatrooms (self);
+  chatroom_manager_get_all (self);
+}
+
 static void
 account_manager_ready_cb (GObject *source_object,
-                          GAsyncResult *result,
-                          gpointer user_data)
+    GAsyncResult *result,
+    gpointer user_data)
 {
   EmpathyChatroomManager *self = EMPATHY_CHATROOM_MANAGER (user_data);
+  EmpathyChatroomManagerPriv *priv = GET_PRIV (self);
   TpAccountManager *manager = TP_ACCOUNT_MANAGER (source_object);
   GError *error = NULL;
+  GFile *file = NULL;
 
   if (!tp_account_manager_prepare_finish (manager, result, &error))
     {
@@ -426,18 +498,37 @@ account_manager_ready_cb (GObject *source_object,
 
   chatroom_manager_get_all (self);
 
+  /* Set up file monitor */
+  file = g_file_new_for_path (priv->file);
+
+  priv->monitor = g_file_monitor (file, 0, NULL, &error);
+  if (priv->monitor == NULL)
+    {
+      DEBUG ("Failed to create file monitor on %s: %s", priv->file,
+          error->message);
+
+      g_error_free (error);
+      goto out;
+    }
+
+  g_signal_connect (priv->monitor, "changed", G_CALLBACK (file_changed_cb),
+      self);
+
 out:
+  tp_clear_object (&file);
   g_object_unref (self);
 }
 
 static GObject *
 empathy_chatroom_manager_constructor (GType type,
-                                      guint n_props,
-                                      GObjectConstructParam *props)
+    guint n_props,
+    GObjectConstructParam *props)
 {
   GObject *obj;
   EmpathyChatroomManager *self;
   EmpathyChatroomManagerPriv *priv;
+  GError *error = NULL;
+  TpDBusDaemon *dbus;
 
   if (chatroom_manager_singleton != NULL)
     return g_object_ref (chatroom_manager_singleton);
@@ -472,6 +563,38 @@ empathy_chatroom_manager_constructor (GType type,
       g_free (dir);
     }
 
+  dbus = tp_dbus_daemon_dup (&error);
+  if (dbus == NULL)
+    {
+      g_warning ("Failed to get TpDBusDaemon: %s", error->message);
+
+      g_error_free (error);
+      return obj;
+    }
+
+  /* Setup a room observer */
+  priv->observer = tp_simple_observer_new (dbus, TRUE,
+      "Empathy.ChatroomManager", TRUE, observe_channels_cb, self, NULL);
+
+  g_object_unref (dbus);
+
+  tp_base_client_take_observer_filter (priv->observer, tp_asv_new (
+      TP_PROP_CHANNEL_CHANNEL_TYPE, G_TYPE_STRING,
+        TP_IFACE_CHANNEL_TYPE_TEXT,
+      TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, G_TYPE_UINT,
+        TP_HANDLE_TYPE_ROOM,
+      NULL));
+
+  tp_base_client_add_connection_features_varargs (priv->observer,
+      TP_CONNECTION_FEATURE_CAPABILITIES, NULL);
+
+  if (!tp_base_client_register (priv->observer, &error))
+    {
+      g_critical ("Failed to register Observer: %s", error->message);
+
+      g_error_free (error);
+    }
+
   return obj;
 }
 
@@ -484,7 +607,8 @@ empathy_chatroom_manager_class_init (EmpathyChatroomManagerClass *klass)
   object_class->constructor = empathy_chatroom_manager_constructor;
   object_class->get_property = empathy_chatroom_manager_get_property;
   object_class->set_property = empathy_chatroom_manager_set_property;
-	object_class->finalize = chatroom_manager_finalize;
+  object_class->dispose = chatroom_manager_dispose;
+  object_class->finalize = chatroom_manager_finalize;
 
   param_spec = g_param_spec_string (
       "file",
@@ -543,7 +667,7 @@ empathy_chatroom_manager_dup_singleton (const gchar *file)
 
 gboolean
 empathy_chatroom_manager_add (EmpathyChatroomManager *manager,
-			     EmpathyChatroom        *chatroom)
+    EmpathyChatroom *chatroom)
 {
   EmpathyChatroomManagerPriv *priv;
 
@@ -571,7 +695,7 @@ empathy_chatroom_manager_add (EmpathyChatroomManager *manager,
 
 static void
 chatroom_manager_remove_link (EmpathyChatroomManager *manager,
-                              GList *l)
+    GList *l)
 {
   EmpathyChatroomManagerPriv *priv;
   EmpathyChatroom *chatroom;
@@ -593,7 +717,7 @@ chatroom_manager_remove_link (EmpathyChatroomManager *manager,
 
 void
 empathy_chatroom_manager_remove (EmpathyChatroomManager *manager,
-                                 EmpathyChatroom        *chatroom)
+    EmpathyChatroom        *chatroom)
 {
   EmpathyChatroomManagerPriv *priv;
   GList *l;
@@ -620,113 +744,86 @@ empathy_chatroom_manager_remove (EmpathyChatroomManager *manager,
 
 EmpathyChatroom *
 empathy_chatroom_manager_find (EmpathyChatroomManager *manager,
-                               TpAccount *account,
-                               const gchar *room)
+    TpAccount *account,
+    const gchar *room)
 {
-	EmpathyChatroomManagerPriv *priv;
-	GList                     *l;
+  EmpathyChatroomManagerPriv *priv;
+  GList *l;
 
-	g_return_val_if_fail (EMPATHY_IS_CHATROOM_MANAGER (manager), NULL);
-	g_return_val_if_fail (room != NULL, NULL);
+  g_return_val_if_fail (EMPATHY_IS_CHATROOM_MANAGER (manager), NULL);
+  g_return_val_if_fail (room != NULL, NULL);
 
-	priv = GET_PRIV (manager);
+  priv = GET_PRIV (manager);
 
-	for (l = priv->chatrooms; l; l = l->next) {
-		EmpathyChatroom *chatroom;
-		TpAccount *this_account;
-		const gchar    *this_room;
+  for (l = priv->chatrooms; l; l = l->next)
+    {
+      EmpathyChatroom *chatroom;
+      TpAccount *this_account;
+      const gchar    *this_room;
 
-		chatroom = l->data;
-		this_account = empathy_chatroom_get_account (chatroom);
-		this_room = empathy_chatroom_get_room (chatroom);
+      chatroom = l->data;
+      this_account = empathy_chatroom_get_account (chatroom);
+      this_room = empathy_chatroom_get_room (chatroom);
 
-		if (this_account && this_room && account == this_account
-				&& strcmp (this_room, room) == 0) {
-			return chatroom;
-		}
-	}
+      if (this_account && this_room && account == this_account
+          && strcmp (this_room, room) == 0)
+        return chatroom;
+    }
 
-	return NULL;
+  return NULL;
 }
 
 EmpathyChatroom *
 empathy_chatroom_manager_ensure_chatroom (EmpathyChatroomManager *manager,
-					  TpAccount *account,
-					  const gchar *room,
-					  const gchar *name)
+    TpAccount *account,
+    const gchar *room,
+    const gchar *name)
 {
-	EmpathyChatroom *chatroom;
+  EmpathyChatroom *chatroom;
 
-	chatroom = empathy_chatroom_manager_find (manager, account, room);
+  chatroom = empathy_chatroom_manager_find (manager, account, room);
 
-	if (chatroom) {
-		return g_object_ref (chatroom);
-	} else {
-		chatroom = empathy_chatroom_new_full (account,
-			room,
-			name,
-			FALSE);
-		empathy_chatroom_manager_add (manager, chatroom);
-		return chatroom;
-	}
+  if (chatroom)
+    {
+      return g_object_ref (chatroom);
+    }
+  else
+    {
+      chatroom = empathy_chatroom_new_full (account,
+        room,
+        name,
+        FALSE);
+      empathy_chatroom_manager_add (manager, chatroom);
+      return chatroom;
+    }
 }
 
 GList *
 empathy_chatroom_manager_get_chatrooms (EmpathyChatroomManager *manager,
-				       TpAccount *account)
+    TpAccount *account)
 {
-	EmpathyChatroomManagerPriv *priv;
-	GList                     *chatrooms, *l;
+  EmpathyChatroomManagerPriv *priv;
+  GList *chatrooms, *l;
 
-	g_return_val_if_fail (EMPATHY_IS_CHATROOM_MANAGER (manager), NULL);
+  g_return_val_if_fail (EMPATHY_IS_CHATROOM_MANAGER (manager), NULL);
 
-	priv = GET_PRIV (manager);
+  priv = GET_PRIV (manager);
 
-	if (!account) {
-		return g_list_copy (priv->chatrooms);
-	}
+  if (!account)
+    return g_list_copy (priv->chatrooms);
 
-	chatrooms = NULL;
-	for (l = priv->chatrooms; l; l = l->next) {
-		EmpathyChatroom *chatroom;
+  chatrooms = NULL;
+  for (l = priv->chatrooms; l; l = l->next)
+    {
+      EmpathyChatroom *chatroom;
 
-		chatroom = l->data;
+      chatroom = l->data;
 
-		if (account == empathy_chatroom_get_account (chatroom)) {
-			chatrooms = g_list_append (chatrooms, chatroom);
-		}
-	}
+      if (account == empathy_chatroom_get_account (chatroom))
+        chatrooms = g_list_append (chatrooms, chatroom);
+    }
 
-	return chatrooms;
-}
-
-guint
-empathy_chatroom_manager_get_count (EmpathyChatroomManager *manager,
-				   TpAccount *account)
-{
-	EmpathyChatroomManagerPriv *priv;
-	GList                     *l;
-	guint                      count = 0;
-
-	g_return_val_if_fail (EMPATHY_IS_CHATROOM_MANAGER (manager), 0);
-
-	priv = GET_PRIV (manager);
-
-	if (!account) {
-		return g_list_length (priv->chatrooms);
-	}
-
-	for (l = priv->chatrooms; l; l = l->next) {
-		EmpathyChatroom *chatroom;
-
-		chatroom = l->data;
-
-		if (account == empathy_chatroom_get_account (chatroom)) {
-			count++;
-		}
-	}
-
-	return count;
+  return chatrooms;
 }
 
 static void
@@ -757,35 +854,50 @@ chatroom_manager_chat_destroyed_cb (EmpathyTpChat *chat,
     }
 }
 
-/* Called by EmpathyChatManager when we are handling a new group chat */
-void
-empathy_chatroom_manager_chat_handled (EmpathyChatroomManager *self,
-    EmpathyTpChat *chat,
-    TpAccount *account)
+static void
+observe_channels_cb (TpSimpleObserver *observer,
+    TpAccount *account,
+    TpConnection *connection,
+    GList *channels,
+    TpChannelDispatchOperation *dispatch_operation,
+    GList *requests,
+    TpObserveChannelsContext *context,
+    gpointer user_data)
 {
-  EmpathyChatroom *chatroom;
-  const gchar *roomname;
+  EmpathyChatroomManager *self = user_data;
+  GList *l;
 
-  roomname = empathy_tp_chat_get_id (chat);
-
-  chatroom = empathy_chatroom_manager_find (self, account, roomname);
-
-  if (chatroom == NULL)
+  for (l = channels; l != NULL; l = g_list_next (l))
     {
-      chatroom = empathy_chatroom_new_full (account, roomname, roomname,
-        FALSE);
-      empathy_chatroom_set_tp_chat (chatroom, chat);
-      empathy_chatroom_manager_add (self, chatroom);
-      g_object_unref (chatroom);
-    }
-  else
-    {
-        empathy_chatroom_set_tp_chat (chatroom, chat);
+      TpChannel *channel = l->data;
+      EmpathyTpChat *tp_chat;
+      const gchar *roomname;
+      EmpathyChatroom *chatroom;
+
+      if (tp_proxy_get_invalidated (channel) != NULL)
+        continue;
+
+      tp_chat = empathy_tp_chat_new (account, channel);
+      roomname = empathy_tp_chat_get_id (tp_chat);
+      chatroom = empathy_chatroom_manager_find (self, account, roomname);
+
+      if (chatroom == NULL)
+        {
+          chatroom = empathy_chatroom_new_full (account, roomname, roomname,
+            FALSE);
+          empathy_chatroom_manager_add (self, chatroom);
+          g_object_unref (chatroom);
+        }
+
+      empathy_chatroom_set_tp_chat (chatroom, tp_chat);
+      g_object_unref (tp_chat);
+
+      /* A TpChat is always destroyed as it only gets unreffed after the channel
+       * has been invalidated in the dispatcher..  */
+      g_signal_connect (tp_chat, "destroy",
+        G_CALLBACK (chatroom_manager_chat_destroyed_cb),
+        self);
     }
 
-  /* A TpChat is always destroyed as it only gets unreffed after the channel
-   * has been invalidated in the dispatcher..  */
-  g_signal_connect (chat, "destroy",
-    G_CALLBACK (chatroom_manager_chat_destroyed_cb),
-    self);
+  tp_observe_channels_context_accept (context);
 }
