@@ -26,6 +26,7 @@
 #include <config.h>
 
 #include <string.h>
+#include <stdio.h>
 
 #include <gtk/gtk.h>
 #include <glib/gi18n.h>
@@ -49,7 +50,7 @@
 
 G_DEFINE_TYPE (EmpathyPreferences, empathy_preferences, GTK_TYPE_DIALOG);
 
-#define GET_PRIV(self) ((EmpathyPreferencesPriv*)((EmpathyPreferences*)self)->priv)
+#define GET_PRIV(self) ((EmpathyPreferencesPriv *)((EmpathyPreferences *) self)->priv)
 
 struct _EmpathyPreferencesPriv {
 	GtkWidget *notebook;
@@ -58,7 +59,9 @@ struct _EmpathyPreferencesPriv {
 	GtkWidget *checkbutton_show_contacts_in_rooms;
 	GtkWidget *combobox_chat_theme;
 	GtkWidget *checkbutton_separate_chat_windows;
+	GtkWidget *checkbutton_events_notif_area;
 	GtkWidget *checkbutton_autoconnect;
+	GtkWidget *checkbutton_logging;
 
 	GtkWidget *checkbutton_sounds_enabled;
 	GtkWidget *checkbutton_sounds_disabled_away;
@@ -84,6 +87,7 @@ struct _EmpathyPreferencesPriv {
 	GSettings *gsettings_notify;
 	GSettings *gsettings_sound;
 	GSettings *gsettings_ui;
+	GSettings *gsettings_logger;
 };
 
 static void     preferences_setup_widgets                (EmpathyPreferences      *preferences);
@@ -98,7 +102,7 @@ static void     preferences_languages_load               (EmpathyPreferences
 static gboolean preferences_languages_load_foreach       (GtkTreeModel           *model,
 							  GtkTreePath            *path,
 							  GtkTreeIter            *iter,
-							  gchar                 **languages);
+							  GList                  *languages);
 static void     preferences_languages_cell_toggled_cb    (GtkCellRendererToggle  *cell,
 							  gchar                  *path_string,
 							  EmpathyPreferences      *preferences);
@@ -221,6 +225,12 @@ preferences_setup_widgets (EmpathyPreferences *preferences)
 			 "active",
 			 G_SETTINGS_BIND_DEFAULT);
 
+	g_settings_bind (priv->gsettings_ui,
+			 EMPATHY_PREFS_UI_EVENTS_NOTIFY_AREA,
+			 priv->checkbutton_events_notif_area,
+			 "active",
+			 G_SETTINGS_BIND_DEFAULT);
+
 	g_settings_bind (priv->gsettings_chat,
 			 EMPATHY_PREFS_CHAT_SHOW_SMILEYS,
 			 priv->checkbutton_show_smileys,
@@ -287,6 +297,12 @@ preferences_setup_widgets (EmpathyPreferences *preferences)
 			 priv->checkbutton_location_reduce_accuracy,
 			 "sensitive",
 			 G_SETTINGS_BIND_GET);
+
+	g_settings_bind (priv->gsettings_logger,
+			 EMPATHY_PREFS_LOGGER_ENABLED,
+			 priv->checkbutton_logging,
+			 "active",
+			 G_SETTINGS_BIND_DEFAULT);
 }
 
 static void
@@ -456,9 +472,6 @@ preferences_languages_add (EmpathyPreferences *preferences)
 
 	codes = empathy_spell_get_language_codes ();
 
-	g_settings_set_boolean (priv->gsettings_chat,
-				EMPATHY_PREFS_CHAT_SPELL_CHECKER_ENABLED,
-				codes != NULL);
 	if (!codes) {
 		gtk_widget_set_sensitive (priv->treeview_spell_checker, FALSE);
 	}
@@ -554,37 +567,34 @@ preferences_languages_load (EmpathyPreferences *preferences)
 	EmpathyPreferencesPriv *priv = GET_PRIV (preferences);
 	GtkTreeView   *view;
 	GtkTreeModel  *model;
-	gchar         *value;
-	gchar        **vlanguages;
+	GList         *enabled_codes;
 
-	value = g_settings_get_string (priv->gsettings_chat,
-				       EMPATHY_PREFS_CHAT_SPELL_CHECKER_LANGUAGES);
+	enabled_codes = empathy_spell_get_enabled_language_codes ();
 
-	if (value == NULL)
+	g_settings_set_boolean (priv->gsettings_chat,
+				EMPATHY_PREFS_CHAT_SPELL_CHECKER_ENABLED,
+				enabled_codes != NULL);
+
+	if (enabled_codes == NULL)
 		return;
 
-	vlanguages = g_strsplit (value, ",", -1);
-	g_free (value);
-
 	view = GTK_TREE_VIEW (priv->treeview_spell_checker);
 	model = gtk_tree_view_get_model (view);
 
 	gtk_tree_model_foreach (model,
 				(GtkTreeModelForeachFunc) preferences_languages_load_foreach,
-				vlanguages);
+				enabled_codes);
 
-	g_strfreev (vlanguages);
+	g_list_free (enabled_codes);
 }
 
 static gboolean
 preferences_languages_load_foreach (GtkTreeModel  *model,
 				    GtkTreePath   *path,
 				    GtkTreeIter   *iter,
-				    gchar        **languages)
+				    GList         *languages)
 {
 	gchar    *code;
-	gchar    *lang;
-	gint      i;
 	gboolean  found = FALSE;
 
 	if (!languages) {
@@ -596,10 +606,8 @@ preferences_languages_load_foreach (GtkTreeModel  *model,
 		return FALSE;
 	}
 
-	for (i = 0, lang = languages[i]; lang; lang = languages[++i]) {
-		if (!tp_strdiff (lang, code)) {
-			found = TRUE;
-		}
+	if (g_list_find_custom (languages, code, (GCompareFunc) strcmp)) {
+		found = TRUE;
 	}
 
 	g_free (code);
@@ -830,6 +838,7 @@ empathy_preferences_finalize (GObject *self)
 	g_object_unref (priv->gsettings_notify);
 	g_object_unref (priv->gsettings_sound);
 	g_object_unref (priv->gsettings_ui);
+	g_object_unref (priv->gsettings_logger);
 
 	G_OBJECT_CLASS (empathy_preferences_parent_class)->finalize (self);
 }
@@ -868,7 +877,6 @@ empathy_preferences_init (EmpathyPreferences *preferences)
 	gtk_window_set_position (GTK_WINDOW (preferences),
 				 GTK_WIN_POS_CENTER_ON_PARENT);
 	gtk_window_set_icon_name (GTK_WINDOW (preferences), "gtk-preferences");
-	gtk_dialog_set_has_separator (GTK_DIALOG (preferences), FALSE);
 
 	filename = empathy_file_lookup ("empathy-preferences.ui", "src");
 	gui = empathy_builder_get_file (filename,
@@ -877,7 +885,9 @@ empathy_preferences_init (EmpathyPreferences *preferences)
 		"checkbutton_show_contacts_in_rooms", &priv->checkbutton_show_contacts_in_rooms,
 		"combobox_chat_theme", &priv->combobox_chat_theme,
 		"checkbutton_separate_chat_windows", &priv->checkbutton_separate_chat_windows,
+		"checkbutton_events_notif_area", &priv->checkbutton_events_notif_area,
 		"checkbutton_autoconnect", &priv->checkbutton_autoconnect,
+		"checkbutton_logging", &priv->checkbutton_logging,
 		"checkbutton_notifications_enabled", &priv->checkbutton_notifications_enabled,
 		"checkbutton_notifications_disabled_away", &priv->checkbutton_notifications_disabled_away,
 		"checkbutton_notifications_focus", &priv->checkbutton_notifications_focus,
@@ -906,6 +916,7 @@ empathy_preferences_init (EmpathyPreferences *preferences)
 	priv->gsettings_notify = g_settings_new (EMPATHY_PREFS_NOTIFICATIONS_SCHEMA);
 	priv->gsettings_sound = g_settings_new (EMPATHY_PREFS_SOUNDS_SCHEMA);
 	priv->gsettings_ui = g_settings_new (EMPATHY_PREFS_UI_SCHEMA);
+	priv->gsettings_logger = g_settings_new (EMPATHY_PREFS_LOGGER_SCHEMA);
 
 	preferences_themes_setup (preferences);
 
@@ -924,7 +935,7 @@ empathy_preferences_init (EmpathyPreferences *preferences)
 	}
 
 	page = gtk_notebook_get_nth_page (GTK_NOTEBOOK (priv->notebook), 3);
-#if HAVE_GEOCLUE
+#ifdef HAVE_GEOCLUE
 	gtk_widget_show (page);
 #else
 	gtk_widget_hide (page);
